from .Constants import *
from can import Message


class AdroitDrive:
    def __init__(self, address):
        self.addr = address
        self.position = 0
        self.velocity = 0
        self.effort = 0
        self.motor_current = 0
        self.bus_voltage = 0
        self.bus_current = 0
        self.temperature = 0
        self.max_velocity = DEFAULT_DRIVE_VEL
        self.max_effort = 0
        self.max_motor_current = MAX_DRIVE_CURRENT
        self.state = NO_STATE

    def set_max_velocity(self, velocity):
        self.max_velocity = velocity

    def set_max_effort(self, effort):
        self.max_effort = effort

    def set_max_motor_current(self, motor_current):
        self.max_motor_current = motor_current

    def get_state_cmd(self, state):
        if state < NO_STATE or state > NUM_STATES:
            raise ValueError("Invalid state {}".format(state))
        return Message(
            arbitration_id=(STATE_CHANGE_CMD << 7) | (self.addr & 0x7F),
            is_extended_id=False,
            data=[state],
        )

    def get_control_cmd(
        self, position,
    ):
        position_fp = int(position * INT16_MAX / POSITION_CONV)
        velocity_fp = int(self.max_velocity * INT16_MAX / VELOCITY_CONV)
        effort_fp = int(self.max_effort * INT16_MAX / EFFORT_CONV)
        motor_current_fp = int(self.max_motor_current * INT16_MAX / CURRENT_CONV)
        return Message(
            arbitration_id=(CONTROL_CMD << 7) | (self.addr & 0x7F),
            is_extended_id=False,
            data=[
                position_fp & 0xFF,
                (position_fp >> 8) & 0xFF,
                velocity_fp & 0xFF,
                (velocity_fp >> 8) & 0xFF,
                effort_fp & 0xFF,
                (effort_fp >> 8) & 0xFF,
                motor_current_fp & 0xFF,
                (motor_current_fp >> 8) & 0xFF,
            ],
        )

    def get_impedance_cmd(self, inertia, damping, stiffness):
        # INT16_MAX may need to be cast to a float
        inertia_fp = int(inertia * INT16_MAX / INERTIA_CONV)
        damping_fp = int(damping * INT16_MAX / DAMPING_CONV)
        stiffness_fp = int(stiffness * INT16_MAX / STIFFNESS_CONV)
        return Message(
            arbitration_id=(IMPEDANCE_CMD << 7) | (self.addr & 0x7F),
            is_extended_id=False,
            data=[
                inertia_fp & 0xFF,
                (inertia_fp >> 8) & 0xFF,
                damping_fp & 0xFF,
                (damping_fp >> 8) & 0xFF,
                stiffness_fp & 0xFF,
                (stiffness_fp >> 8) & 0xFF,
            ],
        )

    def process_telem(self, msg):
        msg_id = (msg.arbitration_id >> 7) & 0x0F
        if msg_id == ERROR_TELEM:
            self.process_error_telem(msg)
        elif msg_id == HIGH_SPEED_TELEM:
            self.process_hs_telem(msg)
        elif msg_id == MEDIUM_SPEED_TELEM:
            self.process_ms_telem(msg)
        elif msg_id == LOW_SPEED_TELEM:
            self.process_ls_telem(msg)
        elif msg_id == DEBUG_TELEM:
            self.process_debug_telem(msg)
        elif msg_id == PARAMETER_TELEM:
            self.process_parameter_telem(msg)
        elif msg_id == STATUS_TELEM:
            self.process_status_telem(msg)
        elif msg_id == CONTROL_CMD:
            self.process_control_cmd(msg)

    def process_error_telem(self, msg):
        print("Recieved error message:")
        print(msg)

    def process_hs_telem(self, msg):
        if msg.dlc == 8:
            position_fp = int.from_bytes(msg.data[:2], 'little', signed=True)
            velocity_fp = int.from_bytes(msg.data[2:4], 'little', signed=True)
            effort_fp = int.from_bytes(msg.data[4:6], 'little', signed=True)
            motor_current_fp = int.from_bytes(msg.data[6:], 'little', signed=True)
            self.position = position_fp / INT16_MAX * POSITION_CONV
            self.velocity = velocity_fp / INT16_MAX * VELOCITY_CONV
            self.effort = effort_fp / INT16_MAX * EFFORT_CONV
            self.motor_current = motor_current_fp / INT16_MAX * CURRENT_CONV

    def process_ms_telem(self, msg):
        pass

    def process_ls_telem(self, msg):
        if msg.dlc == 6:
            bus_voltage_fp = int.from_bytes(msg.data[:2], 'little', signed=True)
            bus_current_fp = int.from_bytes(msg.data[2:4], 'little', signed=True)
            temperature_fp = int.from_bytes(msg.data[4:], 'little', signed=True)
            self.bus_voltage = bus_voltage_fp / INT16_MAX * VOLTAGE_CONV
            self.bus_current = bus_current_fp / INT16_MAX * CURRENT_CONV
            self.temperature = temperature_fp / INT16_MAX * TEMPERATURE_CONV

    def process_debug_telem(self, msg):
        print("Recieved debug message:")
        print(msg)

    def process_parameter_telem(self, msg):
        print("Recieved parameter message:")
        print(msg)

    def process_status_telem(self, msg):
        if msg.dlc == 1:
            self.state = msg.data[0]

    def process_control_cmd(self, msg):
        print("Recieved control message:")
        print(msg)



from math import pi

INT16_MAX = 0x7FFF

INIT_WAIT_TIME = 1000000

MAX_DRIVE_CURRENT = 64
DEFAULT_DRIVE_VEL = 0.2618

BROADCAST_ADDR = 0

POSITION_CONV = 2.0 * pi
VELOCITY_CONV = 8.0
EFFORT_CONV = 256.0
CURRENT_CONV = 64.0
VOLTAGE_CONV = 64.0
TEMPERATURE_CONV = 128.0
INERTIA_CONV = 32.0
DAMPING_CONV = 128.0
STIFFNESS_CONV = 1024.0

NO_STATE = -1
INIT_STATE = 0
PROGRAM_STATE = 1
STARTUP_STATE = 2
RUN_STATE = 3
HOLD_STATE = 4
SLEEP_STATE = 5
FAULT_STATE = 6
NUM_STATES = 7

STATE_CHANGE_CMD = 0
ERROR_TELEM = 1
HIGH_SPEED_TELEM = 3
CONTROL_CMD = 4
MEDIUM_SPEED_TELEM = 5
IMPEDANCE_CMD = 6
LOW_SPEED_TELEM = 7
DEBUG_TELEM = 9
CURRENT_CMD = 10
PARAMETER_TELEM = 11
PARAMETER_CMD = 12
STATUS_CMD = 14
STATUS_TELEM = 15



import setuptools

with open("README.md") as f:
    long_description = f.read()

setuptools.setup(
    name="AdroitComs",
    version="0.0.1",
    author="Cameron Devine",
    author_email="camdev@uw.edu",
    description="A package for communicating with HDT Adroit robots over CAN.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=[
        'python-can',
        'numpy',
    ]
)



from .Drive import AdroitDrive
from .Constants import *
import can
import numpy as np
from threading import Thread
from .Util import usleep

__all__ = ["AdroitComs"]


class AdroitComs:
    def __init__(self, drives, *args, **kwargs):
        self.bus = can.Bus(*args, **kwargs)
        self.drives = [AdroitDrive(i + 1) for i in range(drives)]
        self.broadcast = AdroitDrive(BROADCAST_ADDR)
        self.tx = []
        self.rx = []
        self.threads = []
        self.run = True
        self.start(self.io_handler)
        self.start(self.telem_handler)

    def start(self, target):
        thread = Thread(target=target)
        thread.daemon = True
        thread.start()
        self.threads.append(thread)

    def send(self, msg):
        self.tx.append(msg)

    def get_drives(self, drive):
        if drive is None:
            return self.drives
        elif drive == "broadcast":
            return [self.broadcast]
        elif isinstance(drive, int):
            if drive >= len(self.drives):
                raise ValueError("drive is larger than the number of conncected drives")
            return [self.drives[drive]]
        else:
            raise ValueError(
                "drive must be either, None, 'broadcast', or a drive number."
            )

    def reset(self):
        self.send(self.broadcast.get_state_cmd(INIT_STATE))
        while len(self.tx):
            pass
        usleep(INIT_WAIT_TIME)
        for drive in self.drives:
            self.send(drive.get_state_cmd(STARTUP_STATE))
        while len(self.tx):
            pass
        usleep(2 * INIT_WAIT_TIME)
        return np.all((STARTUP_STATE < self.state) & (self.state < FAULT_STATE))

    def send_state_cmd(self, state, drive=None):
        for d in self.get_drives(drive):
            self.send(d.get_state_cmd(state))

    def send_control_cmd(self, position, drive=None):
        for d, pos, in zip(self.get_drives(drive), position):
            self.send(d.get_control_cmd(pos))

    def set_max(self, setting, value, drive):
        drives = self.get_drives(drive)
        if isinstance(value, (float, int)):
            value = len(drives) * [value]
        for d, v in zip(drives, value):
            getattr(d, "set_max_" + setting)(v)

    def set_max_velocity(self, velocity, drive=None):
        self.set_max("velocity", velocity, drive)

    def set_max_effort(self, effort, drive=None):
        self.set_max("effort", effort, drive)

    def set_max_motor_current(self, motor_current, drive=None):
        self.set_max("motor_current", motor_current, drive)

    def send_impedance_cmd(self, inertia, damping, stiffness, drive=None):
        for d in self.get_drives(drive):
            self.send(d.get_impedance_cmd(inertia, damping, stiffness))

    def shutdown(self):
        self.run = False

    def io_handler(self):
        while self.run:
            if len(self.tx):
                self.bus.send(self.tx.pop(0), 0)
        self.bus.shutdown()

    def telem_handler(self):
        while self.run:
            recv = self.bus.recv(1)
            if recv is not None:
                self.process_telem(recv)

    def process_telem(self, msg):
        addr = msg.arbitration_id & 0x7F
        if addr == 0:
            self.broadcast.process_telem(msg)
        elif addr <= len(self.drives):
            self.drives[addr - 1].process_telem(msg)

    def get_telem(self, attrib):
        telem = np.zeros((len(self.drives,)))
        for i, drive in enumerate(self.drives):
            telem[i] = getattr(drive, attrib)
        return telem

    @property
    def position(self):
        return self.get_telem("position")

    @property
    def velocity(self):
        return self.get_telem("velocity")

    @property
    def effort(self):
        return self.get_telem("effort")

    @property
    def motor_current(self):
        return self.get_telem("motor_current")

    @property
    def bus_voltage(self):
        return self.get_telem("bus_voltage")

    @property
    def bus_current(self):
        return self.get_telem("bus_current")

    @property
    def temperature(self):
        return self.get_telem("temperature")

    @property
    def state(self):
        return self.get_telem("state")
